#include <time.h>
#include <assert.h>
#include <stdio.h>
#include <sstream>
using namespace std;
#include <NTL/ZZX.h>
using namespace NTL;
#include "det.h"

long long nanoseconds(void)
{
  struct timespec ts;
  clock_gettime(CLOCK_MONOTONIC,&ts);
  return ts.tv_nsec+ts.tv_sec*(long long) 1000000000;
}

#define SCALE 10
#define NMAX 1024
#define BITSMAX 128

int ntlcheck = 1;

int main()
{
  mpz_t det;
  mpz_init(det);
  mpz_t absdet;
  mpz_init(absdet);
  mpz_t check;
  mpz_init(check);
  mpz_t f[NMAX];
  for (long long i = 0;i < NMAX;++i)
    mpz_init(f[i]);
  mpz_t fneg;
  mpz_init(fneg);

  for (long long n = 1;n <= NMAX;n *= 2) {
    ZZX Phi;
    SetCoeff(Phi,0,1);
    SetCoeff(Phi,n,1);

    for (long long fbits = 1;fbits <= BITSMAX;fbits += 1+fbits/4) {
      for (long long real = 0;real < 2;++real) {
        long long cyclo2powerns = 0; long long cyclo2powercalls = 0;
        long long ntlns = 0; long long ntlcalls = 0;
        
        for (long long loop = 0;loop < NMAX*SCALE;loop += n) {

          ZZX F;
          for (long long i = 0;i < n;++i) {
            ZZ r = RandomBits_ZZ(fbits);
            if (RandomBits_ZZ(1) > 0) r = -r;
            if (real) {
              if (i*2 == n) r = 0;
              if (i > n-i) r = -coeff(F,n-i);
            }
            SetCoeff(F,i,r);
          }
            
          for (long long i = 0;i < n;++i) {
            stringstream ss;
            ss << coeff(F,i);
            mpz_set_str(f[i],ss.str().c_str(),10);
          }

          if (real) {
            cyclo2powerns -= nanoseconds();
            cyclo2power_det_signed_real(det,f,n);
            cyclo2powerns += nanoseconds(); ++cyclo2powercalls;
            cyclo2power_det_abs_real(absdet,f,n);
            if (n > 1) {
              mpz_mul(det,det,det);
              mpz_mul(absdet,absdet,absdet);
              assert(mpz_cmp(det,absdet) == 0);
              // XXX: with more work can check if det had the right sign
            }
          } else {
            cyclo2powerns -= nanoseconds();
            cyclo2power_det_signed(det,f,n);
            cyclo2powerns += nanoseconds(); ++cyclo2powercalls;
            cyclo2power_det_abs(absdet,f,n);
          }

          if (ntlcheck) {
            ntlns -= nanoseconds();
            ZZ ntldet = resultant(Phi,F);
            ntlns += nanoseconds(); ++ntlcalls;
  
            stringstream ss;
            ss << ntldet;
  
            mpz_set_str(check,ss.str().c_str(),10);
            mpz_sub(check,check,det);
            assert(mpz_sgn(check) == 0);
  
            mpz_set_str(check,ss.str().c_str(),10);
            mpz_abs(check,check);
            mpz_sub(check,check,absdet);
            assert(mpz_sgn(check) == 0);
          }
        }

        if (ntlcheck) {
          assert(ntlcalls == cyclo2powercalls);
          printf("n %lld fbits %lld real %lld ntlns %lld cyclo2powerns %lld calls %lld ratio %lld\n",n,fbits,real,ntlns/ntlcalls,cyclo2powerns/ntlcalls,ntlcalls,ntlns/cyclo2powerns);
        } else {
          printf("n %lld fbits %lld real %lld cyclo2powerns %lld calls %lld\n",n,fbits,real,cyclo2powerns/cyclo2powercalls,cyclo2powercalls);
        }
        fflush(stdout);
      }
    }
  }

  mpz_clear(det);
  mpz_clear(absdet);
  mpz_clear(check);
  for (long long i = 0;i < NMAX;++i)
    mpz_clear(f[i]);
  mpz_clear(fneg);
}
